/**
   @Author:
      Date: <%b> <%d>, <%Y>
   Teacher:
       Lab:
      Misc:
  */

import java.util.ArrayList;

public class UnboundedEnv extends SquareEnvironment
{
    // Instance Variables: Encapsulated data for each UnboundedEnv object
    private ArrayList objectList;   // list of Locatable objects in environment

    public UnboundedEnv()
    {
        // Construct and initialize inherited attributes.
        super();

        objectList = new ArrayList();
    }

    public int numRows()
    {
        return -1;
    }

    public int numCols()
    {
        return -1;
    }

    public boolean isValid(Location loc)
    {
        // All non-null locations are valid in an unbounded environment.
        return loc != null;
    }

    public int numObjects()
    {
        return objectList.size();
    }

    public Locatable[] allObjects()
    {
        Locatable[] objectArray = new Locatable[objectList.size()];

        // Put all the environment objects in the list.
        for ( int index = 0; index < objectList.size(); index++ )
        {
            objectArray[index] = (Locatable) objectList.get(index);
        }

        return objectArray;
    }

    public boolean isEmpty(Location loc)
    {
        return (objectAt(loc) == null);
    }

    public Locatable objectAt(Location loc)
    {
        int index = indexOf(loc);
        if ( index == -1 )
            return null;

        return (Locatable) objectList.get(index);
    }

    public String toString()
    {
        Locatable[] theObjects = allObjects();
        String s = "Environment contains " + numObjects() + " objects: ";
        for ( int index = 0; index < theObjects.length; index++ )
            s += theObjects[index].toString() + " ";
        return s;
    }

    public void add(Locatable obj)
    {
        // Check precondition.  Location should be empty.
        Location loc = obj.location();
        if ( ! isEmpty(loc) )
            throw new IllegalArgumentException("Location " + loc +
                                    " is not a valid empty location");

        // Add object to the environment.
        objectList.add(obj);
    }

    public void remove(Locatable obj)
    {
        // Find the index of the object to remove.
        int index = indexOf(obj.location());
        if ( index == -1 )
            throw new IllegalArgumentException("Cannot remove " +
                                               obj + "; not there");

        // Remove the object.
        objectList.remove(index);
    }

    public void recordMove(Locatable obj, Location oldLoc)
    {
        int objectsAtOldLoc = 0;
        int objectsAtNewLoc = 0;

        // Look through the list to find how many objects are at old
        // and new locations.
        Location newLoc = obj.location();
        for ( int index = 0; index < objectList.size(); index++ )
        {
            Locatable thisObj = (Locatable) objectList.get(index);
            if ( thisObj.location().equals(oldLoc) )
                objectsAtOldLoc++;
            if ( thisObj.location().equals(newLoc) )
                objectsAtNewLoc++;
        }

        // There should be one object at newLoc.  If oldLoc equals
        // newLoc, there should be one at oldLoc; otherwise, there
        // should be none.
        if ( ! ( objectsAtNewLoc == 1 &&
                 ( oldLoc.equals(newLoc) || objectsAtOldLoc == 0 ) ) )
        {
            throw new IllegalArgumentException("Precondition violation moving "
                + obj + " from " + oldLoc);
        }
    }

    protected int indexOf(Location loc)
    {
        // Look through the list to find the object at the given location.
        for ( int index = 0; index < objectList.size(); index++ )
        {
            Locatable obj = (Locatable) objectList.get(index);
            if ( obj.location().equals(loc) )
            {
                // Found the object -- return its index.
                return index;
            }
        }

        // No such object found.
        return -1;
    }
}
